# -*- coding: utf-8 -*-
# pylint: disable=invalid-name,too-many-instance-attributes, too-many-arguments
"""
Copyright 2022 Paul A Beltyukov
Copyright 2019 Paul A Beltyukov
Copyright 2015 Roger R Labbe Jr.

FilterPy library.
http://github.com/rlabbe/filterpy

Documentation at:
https://filterpy.readthedocs.org

Supporting book at:
https://github.com/rlabbe/Kalman-and-Bayesian-Filters-in-Python

This is licensed under an MIT license. See the readme.MD file
for more information.
"""
import numpy as np
from numpy import dot, outer
from scipy.stats import chi2
from UDEKF import mwgs, UDExtendedKalmanFilter


class UDExtendedKalmanHinfFilterPrior2(UDExtendedKalmanFilter):

    """
    Implements an UD modification of extended Kalman/Hinfinity filter (EKHF)
    with prior resuduals used for Hinfinity correction. You are responsible
    for setting the various state variables to reasonable values;
    the defaults will  not give you a functional filter.

    You will have to set the following attributes after constructing this
    object for the filter to perform properly. Please note that there are
    various checks in place to ensure that you have made everything the
    'correct' size. However, it is possible to provide incorrectly sized
    arrays such that the linear algebra can not perform an operation.
    It can also fail silently - you can end up with matrices of a size that
    allows the linear algebra to work, but are the wrong shape for the problem
    you are trying to solve.

    Parameters
    ----------

    dim_x : int
        Number of state variables for the Kalman filter. For example, if
        you are tracking the position and velocity of an object in two
        dimensions, dim_x would be 4.

        This is used to set the default size of P, Q, and u

    dim_z : int
        Number of of measurement inputs. For example, if the sensor
        provides you with position in (x,y), dim_z would be 2.

    Attributes
    ----------
    x : numpy.array(dim_x, 1)
        State estimate vector

    P : numpy.array(dim_x, dim_x)
        Covariance matrix

    x_prior : numpy.array(dim_x, 1)
        Prior (predicted) state estimate. The *_prior and *_post attributes
        are for convienence; they store the  prior and posterior of the
        current epoch. Read Only.

    P_prior : numpy.array(dim_x, dim_x)
        Prior (predicted) state covariance matrix. Read Only.

    x_post : numpy.array(dim_x, 1)
        Posterior (updated) state estimate. Read Only.

    P_post : numpy.array(dim_x, dim_x)
        Posterior (updated) state covariance matrix. Read Only.

    R : numpy.array(dim_z, dim_z)
        Measurement noise matrix

    Q : numpy.array(dim_x, dim_x)
        Process noise matrix

    F : numpy.array()
        State Transition matrix

    H : numpy.array(dim_x, dim_x)
        Measurement function

    y : numpy.array
        Residual of the update step. Read only.

    K : numpy.array(dim_x, dim_z)
        Kalman gain of the update step. Read only.

    S :  numpy.array
        Systen uncertaintly projected to measurement space. Read only.

    z : ndarray
        Last measurement used in update(). Read only.

    log_likelihood : float
        log-likelihood of the last measurement. Read only.

    likelihood : float
        likelihood of last measurment. Read only.

        Computed from the log-likelihood. The log-likelihood can be very
        small,  meaning a large negative value such as -28000. Taking the
        exp() of that results in 0.0, which can break typical algorithms
        which multiply by this value, so by default we always return a
        number >= sys.float_info.min.

    mahalanobis : float
        mahalanobis distance of the innovation. E.g. 3 means measurement
        was 3 standard deviations away from the predicted value.

        Read only.
    """

    def __init__(self, dim_x, dim_z, dim_u=0, alpha = 0.001):
        UDExtendedKalmanFilter.__init__(self, dim_x, dim_z, dim_u)
        self.beta_1 = chi2.ppf(1.0 - alpha, 1)

    def _scalar_update(self, nu, h, r):
        """Joseph/Hinfinity scalar update using prior errors

        Parameters
        ----------

        axis_residual : function which returns current axis residual
                        returns scalar, float.

        axis_hjacobian : function which returns current axis HJacobian row
                         returns np.array, float.

        r : scalar, float, current axis state disp
        """
        u, d, n = self.U, self.D, self.dim_x

        f = h.dot(u)
        v = d * f
        c = f.dot(v)
        s = c + r

        a = (nu * nu) / self.beta_1 - s
        if a > 0.0:
            #Divergence detected, H-infinity correction needed
            k = a / c + 1.0
            d *= k
            v *= k
            s = k * c + r

        K = u.dot(v / s).reshape((n, 1))

        WW = np.concatenate((outer(K, f) - u, K), axis = 1)
        DD = np.concatenate((d, np.array([r])))

        self.U, self.D = mwgs(WW, DD)
        self.x += (K*nu).reshape(self.x.shape)


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
    Copyright 2022 anonimous <shkolnick-kun@gmail.com> and contributors.

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing,
    software distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.

    See the License for the specific language governing permissions
"""
import time
import matplotlib.pyplot as plt
import numpy as np

from ab_tests import *
from case1    import *

from yaflpy import AdaptiveRobustJoseph           as A
from yaflpy import UnscentedAdaptiveRobustBierman as B

#------------------------------------------------------------------------------
N = 10000
STD = 100.

#------------------------------------------------------------------------------
clear, noisy, t = case_data(N, STD)

#------------------------------------------------------------------------------
a = case_ekf(A, STD)
b = case_ukf(B, STD)
b.chi2 = a.chi2
#------------------------------------------------------------------------------
start = time.time()

rpa,rua,xa, rpb,rub,xb, nup,ndp, nuq,ndq, nur,ndr, nx, ny = yafl_ab_test(a, b, noisy)

end = time.time()
print(end - start)

#------------------------------------------------------------------------------
plt.plot(nup)
plt.show()
plt.plot(ndp)
plt.show()

plt.plot(nuq)
plt.show()
plt.plot(ndq)
plt.show()

plt.plot(nur)
plt.show()
plt.plot(ndr)
plt.show()

plt.plot(nx)
plt.show()

plt.plot(ny)
plt.show()

plt.plot(noisy[:,0], noisy[:,1], xa[:,0], xa[:,2])
plt.show()

plt.plot(clear[:,0], clear[:,1], xa[:,0], xa[:,2])
plt.show()


# -*- coding: utf-8 -*-
# pylint: disable=invalid-name,too-many-instance-attributes, too-many-arguments
"""
Copyright 2022 Paul A Beltyukov
Copyright 2019 Paul A Beltyukov
Copyright 2015 Roger R Labbe Jr.

FilterPy library.
http://github.com/rlabbe/filterpy

Documentation at:
https://filterpy.readthedocs.org

Supporting book at:
https://github.com/rlabbe/Kalman-and-Bayesian-Filters-in-Python

This is licensed under an MIT license. See the readme.MD file
for more information.
"""
from copy import deepcopy
from math import log, exp, sqrt
import sys
import numpy as np
from numpy import dot, outer, eye, zeros, ones, diag
import scipy.linalg as linalg
from filterpy.stats import logpdf
from filterpy.common import pretty_str, reshape_z
'''
UDU decomposition:
P = U * diag(D) * U^T
'''
def udu(p):

    if 2 != len(p.shape):
        return None

    if p.shape[0] != p.shape[1]:
        return None

    n = p.shape[0]

    u = zeros((n, n))
    d = zeros((n))

    d[n-1] = p[n-1,n-1]
    u[:,n-1]   = p[:,n-1] / d[n-1]

    for j in range(n-2, -1, -1):
        dd = d[j+1:]
        c = dd * u[j,j+1:] #dd is meant to be diag(d[j+1:])
        d[j] = p[j,j] - np.dot(u[j,j+1:].T, c)

        if d[j] == 0:
            return None

        for i in range(j, -1, -1):
            c = dd * u[j,j+1:]
            u[i,j] = (p[i,j] - np.dot(u[i,j+1:].T, c))/d[j]

    return u, d
'''
MWGS update:
U * diag(D) * U^T = w * diag(d) * w^T

Params:
w - is n*k float full rank
d - is k*None float
where k>n
return:
u - is n*n float upper triangular
D - id n*None float
'''
def mwgs(w,d):

    if 1 != len(d.shape):
        return None

    if 2 != len(w.shape):
        return None

    if w.shape[1] != d.shape[0]:
        return None

    if w.shape[0] >= d.shape[0]:
        return None

    n = w.shape[0]

    u = np.eye(n)
    D = np.zeros((n))

    for i in range(n-1, -1, -1):
        c = w[i,:] * d
        D[i] = np.dot(w[i,:], c)

        if D[i] <= 1e-15:
            # How about partial reset heuristics here?
            D[i] = 1e-15
            for j in range(0, i):
                u[j,i] = 0
            continue

        dd = c/D[i]

        for j in range(0, i):
            u[j,i]  = np.dot(dd, w[j,:])
            w[j,:] -= u[j,i] * w[i,:]

    return u, D


class UDExtendedKalmanFilter(object):

    """ Implements an UD modification of extended Kalman filter (EKF).
    You are responsible for setting the various state variables to
    reasonable values; the defaults will  not give you a functional filter.

    You will have to set the following attributes after constructing this
    object for the filter to perform properly. Please note that there are
    various checks in place to ensure that you have made everything the
    'correct' size. However, it is possible to provide incorrectly sized
    arrays such that the linear algebra can not perform an operation.
    It can also fail silently - you can end up with matrices of a size that
    allows the linear algebra to work, but are the wrong shape for the problem
    you are trying to solve.

    Parameters
    ----------

    dim_x : int
        Number of state variables for the Kalman filter. For example, if
        you are tracking the position and velocity of an object in two
        dimensions, dim_x would be 4.

        This is used to set the default size of P, Q, and u

    dim_z : int
        Number of of measurement inputs. For example, if the sensor
        provides you with position in (x,y), dim_z would be 2.

    Attributes
    ----------
    x : numpy.array(dim_x, 1)
        State estimate vector

    P : numpy.array(dim_x, dim_x)
        Covariance matrix

    x_prior : numpy.array(dim_x, 1)
        Prior (predicted) state estimate. The *_prior and *_post attributes
        are for convienence; they store the  prior and posterior of the
        current epoch. Read Only.

    P_prior : numpy.array(dim_x, dim_x)
        Prior (predicted) state covariance matrix. Read Only.

    x_post : numpy.array(dim_x, 1)
        Posterior (updated) state estimate. Read Only.

    P_post : numpy.array(dim_x, dim_x)
        Posterior (updated) state covariance matrix. Read Only.

    R : numpy.array(dim_z, dim_z)
        Measurement noise matrix

    Q : numpy.array(dim_x, dim_x)
        Process noise matrix

    F : numpy.array()
        State Transition matrix

    H : numpy.array(dim_x, dim_x)
        Measurement function

    y : numpy.array
        Residual of the update step. Read only.

    K : numpy.array(dim_x, dim_z)
        Kalman gain of the update step. Read only.

    S :  numpy.array
        Systen uncertaintly projected to measurement space. Read only.

    z : ndarray
        Last measurement used in update(). Read only.

    log_likelihood : float
        log-likelihood of the last measurement. Read only.

    likelihood : float
        likelihood of last measurment. Read only.

        Computed from the log-likelihood. The log-likelihood can be very
        small,  meaning a large negative value such as -28000. Taking the
        exp() of that results in 0.0, which can break typical algorithms
        which multiply by this value, so by default we always return a
        number >= sys.float_info.min.

    mahalanobis : float
        mahalanobis distance of the innovation. E.g. 3 means measurement
        was 3 standard deviations away from the predicted value.

        Read only.
    """

    def __init__(self, dim_x, dim_z, dim_u=0):

        self.dim_x = dim_x
        self.dim_z = dim_z
        self.dim_u = dim_u

        self.x = zeros((dim_x, 1)) # state
        # uncertainty covariance
        self.U = eye(dim_x)
        self.D = ones((dim_x))
        self.B = 0                 # control transition matrix
        self.F = np.eye(dim_x)     # state transition matrix
        # state uncertainty
        self.Dm = eye(dim_z)       #Decorrelation matrix
        self.Ur = eye(dim_z)       #Decorrelation matrix
        self.Dr = ones((dim_z))
        # process uncertainty
        self.Uq = eye(dim_x)
        self.Dq = ones((dim_x))

        z = np.array([None]*self.dim_z)
        self.z = reshape_z(z, self.dim_z, self.x.ndim)

        # residual is computed during the innovation step. We
        # save them so that in case you want to inspect it for various
        # purposes
        self.y = zeros((dim_z, 1)) # residual
        self.S = np.zeros((dim_z, dim_z))   # system uncertainty
        self.SI = np.zeros((dim_z, dim_z))  # inverse system uncertainty

        self._log_likelihood = log(sys.float_info.min)
        self._likelihood = sys.float_info.min
        self._mahalanobis = None

        # these will always be a copy of x,P after predict() is called
        self.x_prior = self.x.copy()
        self.U_prior = self.U.copy()
        self.D_prior = self.D.copy()

        # these will always be a copy of x,P after update() is called
        self.x_post = self.x.copy()
        self.U_post = self.U.copy()
        self.D_post = self.D.copy()

    @property
    def Q(self):
        """ Process uncertainty"""
        return dot(self.Uq, dot(diag(self.Dq), self.Uq.T))

    @Q.setter
    def Q(self, value):
        """ Process uncertainty"""
        self.Uq, self.Dq = udu(value)

    @property
    def P(self):
        """ covariance matrix"""
        return dot(self.U, dot(diag(self.D), self.U.T))

    @property
    def P_prior(self):
        """ covariance matrix of the prior"""
        return dot(self.U_prior, dot(diag(self.D_prior), self.U_prior.T))

    @property
    def P_post(self):
        """ covariance matrix of the posterior"""
        return dot(self.U_post, dot(diag(self.D_post), self.U_post.T))

    @P.setter
    def P(self, value):
        """ covariance matrix"""
        self.U,self.D = udu(value);

    @property
    def R(self):
        """ measurement uncertainty"""
        return dot(self.Ur, dot(diag(self.Dr), self.Ur.T))

    @R.setter
    def R(self, value):
        """ measurement uncertainty"""
        self.Ur, self.Dr = udu(value)
        self.Dm = linalg.inv(self.Ur)

    def predict_x(self, u=0):
        """
        Predicts the next state of X. If you need to
        compute the next state yourself, override this function. You would
        need to do this, for example, if the usual Taylor expansion to
        generate F is not providing accurate results for you.
        """
        self.x = dot(self.F, self.x) + dot(self.B, u)

    def predict(self, u=0):
        """
        Predict next state (prior) using the Kalman filter state propagation
        equations.

        Parameters
        ----------

        u : np.array
            Optional control vector. If non-zero, it is multiplied by B
            to create the control input into the system.
        """

        self.predict_x(u)

        W = np.concatenate((self.Uq, dot(self.F, self.U)), axis=1)
        D = np.concatenate((self.Dq, self.D))
        self.U, self.D = mwgs(W, D)

        # save prior
        self.x_prior = np.copy(self.x)
        self.U_prior = np.copy(self.U)
        self.D_prior = np.copy(self.D)

    def _scalar_update(self, nu, h, r):
        """Joseph scalar update

        Parameters
        ----------

        axis_residual : function which returns current axis residual
                        returns scalar, float.

        axis_hjacobian : function which returns current axis HJacobian row
                         returns np.array, float.

        r : scalar, float, current axis state disp
        """
        u, d, n = self.U, self.D, self.dim_x

        f = h.dot(u)
        v = d * f

        a = r + f.dot(v)
        K = u.dot(v / a).reshape((n, 1))

        WW = np.concatenate((outer(K, f) - u, K), axis = 1)
        DD = np.concatenate((d, np.array([r])))

        self.U, self.D = mwgs(WW, DD)
        self.x += (K*nu).reshape(self.x.shape)

    def update(self, z, HJacobian, Hx, R=None, args=(), hx_args=(),
               residual=np.subtract):
        """ Performs the update innovation of the extended Kalman filter.

        Parameters
        ----------

        z : np.array
            measurement for this step.
            If `None`, posterior is not computed

        HJacobian : function
           function which computes the Jacobian of the H matrix (measurement
           function). Takes state variable (self.x) as input, returns H.

        Hx : function
            function which takes as input the state variable (self.x) along
            with the optional arguments in hx_args, and returns the measurement
            that would correspond to that state.

        R : np.array, scalar, or None
            Optionally provide R to override the measurement noise for this
            one call, otherwise  self.R will be used.

        args : tuple, optional, default (,)
            arguments to be passed into HJacobian after the required state
            variable. for robot localization you might need to pass in
            information about the map and time of day, so you might have
            `args=(map_data, time)`, where the signature of HCacobian will
            be `def HJacobian(x, map, t)`

        hx_args : tuple, optional, default (,)
            arguments to be passed into Hx function after the required state
            variable.

        residual : function (z, z2), optional
            Optional function that computes the residual (difference) between
            the two measurement vectors. If you do not provide this, then the
            built in minus operator will be used. You will normally want to use
            the built in unless your residual computation is nonlinear (for
            example, if they are angles)
        """
        if z is None:
            self.z = np.array([[None]*self.dim_z]).T
            self.x_post = self.x.copy()
            self.P_post = self.P.copy()
            return

        if not isinstance(args, tuple):
            args = (args,)

        if not isinstance(hx_args, tuple):
            hx_args = (hx_args,)

        if R is None:
            Dm = self.Dm
            Dr = self.Dr
        elif np.isscalar(R):
            Dm = eye(self.dim_z)       #Decorrelation matrix
            Dr = ones((self.dim_z)) * R
        else:
            u,d = udu(R);
            Dm = linalg.inv(u)
            Dr = d

        if np.isscalar(z) and self.dim_z == 1:
            z = np.asarray([z], float)

        #The ExtendedKalmanFilter class has self.y, self.S, and self.SI
        #so we have to update them for [partial] compatibility.
        #And yes, this is completely ineffective!!!
        H = HJacobian(self.x, *args)
        hx = Hx(self.x, *hx_args)
        self.y = residual(z, hx)

        self.S = dot(H, dot(self.P, H.T)) + self.R
        self.SI = linalg.inv(self.S)

        #Scalar updates
        dy = dot(Dm, self.y)
        dh = dot(Dm, H)
        for j in range(self.dim_z):
            self._scalar_update(dy[j], dh[j], Dr[j])

        # set to None to force recompute
        self._log_likelihood = None
        self._likelihood = None
        self._mahalanobis = None

        # save measurement and posterior state
        self.z = deepcopy(z)
        self.x_post = self.x.copy()
        self.U_post = self.U.copy()
        self.D_post = self.D.copy()

    def predict_update(self, z, HJacobian, Hx, args=(), hx_args=(), u=0):
        """ Performs the predict/update innovation of the extended Kalman
        filter.

        Parameters
        ----------

        z : np.array
            measurement for this step.
            If `None`, only predict step is perfomed.

        HJacobian : function
           function which computes the Jacobian of the H matrix (measurement
           function). Takes state variable (self.x) as input, along with the
           optional arguments in args, and returns H.

        Hx : function
            function which takes as input the state variable (self.x) along
            with the optional arguments in hx_args, and returns the measurement
            that would correspond to that state.

        args : tuple, optional, default (,)
            arguments to be passed into HJacobian after the required state
            variable.

        hx_args : tuple, optional, default (,)
            arguments to be passed into Hx after the required state
            variable.

        u : np.array or scalar
            optional control vector input to the filter.
        """
        self.predict(u)
        self.update(z, HJacobian, Hx, self.R, args, hx_args, residual=np.subtract)

    @property
    def log_likelihood(self):
        """
        log-likelihood of the last measurement.
        """

        if self._log_likelihood is None:
            self._log_likelihood = logpdf(x=self.y, cov=self.S)
        return self._log_likelihood

    @property
    def likelihood(self):
        """
        Computed from the log-likelihood. The log-likelihood can be very
        small,  meaning a large negative value such as -28000. Taking the
        exp() of that results in 0.0, which can break typical algorithms
        which multiply by this value, so by default we always return a
        number >= sys.float_info.min.
        """
        if self._likelihood is None:
            self._likelihood = exp(self.log_likelihood)
            if self._likelihood == 0:
                self._likelihood = sys.float_info.min
        return self._likelihood

    @property
    def mahalanobis(self):
        """
        Mahalanobis distance of innovation. E.g. 3 means measurement
        was 3 standard deviations away from the predicted value.

        Returns
        -------
        mahalanobis : float
        """
        if self._mahalanobis is None:
            self._mahalanobis = sqrt(float(dot(dot(self.y.T, self.SI), self.y)))
        return self._mahalanobis

    def __repr__(self):
        return '\n'.join([
            'KalmanFilter object',
            pretty_str('x', self.x),
            pretty_str('P', self.P),
            pretty_str('x_prior', self.x_prior),
            pretty_str('P_prior', self.P_prior),
            pretty_str('F', self.F),
            pretty_str('Q', self.Q),
            pretty_str('R', self.R),
            pretty_str('K', self.K),
            pretty_str('y', self.y),
            pretty_str('S', self.S),
            pretty_str('likelihood', self.likelihood),
            pretty_str('log-likelihood', self.log_likelihood),
            pretty_str('mahalanobis', self.mahalanobis)
            ])



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
    Copyright 2022 anonimous <shkolnick-kun@gmail.com> and contributors.

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing,
    software distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.

    See the License for the specific language governing permissions
"""
import h5py
import numpy as np
from numpy.linalg import norm

#------------------------------------------------------------------------------
def yafl_ab_test(a, b, measurements):

    rpa = []
    rua = []
    xa  = []

    rpb = []
    rub = []
    xb  = []

    nup = []
    ndp = []
    nuq = []
    ndq = []
    nur = []
    ndr = []
    nx  = []
    ny  = []

    for z in measurements:
        # Run A
        rpa.append(a.predict())
        rua.append(a.update(z))
        xa.append(a.x.copy())

        # Run B
        rpb.append(b.predict())
        rub.append(b.update(z))
        xb.append(b.x.copy())

        nx.append(2. * norm(a.x - b.x) / norm(a.x + b.x))
        ny.append(2. * norm(a.y - b.y) / norm(a.y + b.y))

        nup.append(2. * norm(a.Up - b.Up) / norm(a.Up + b.Up))
        ndp.append(2. * norm(a.Dp - b.Dp) / norm(a.Dp + b.Dp))

        nuq.append(2. * norm(a.Uq - b.Uq) / norm(a.Uq + b.Uq))
        ndq.append(2. * norm(a.Dq - b.Dq) / norm(a.Dq + b.Dq))

        nur.append(2. * norm(a.Ur - b.Ur) / norm(a.Ur + b.Ur))
        ndr.append(2. * norm(a.Dr - b.Dr) / norm(a.Dr + b.Dr))

    rpa = np.array(rpa)
    rua = np.array(rua)
    xa  = np.array(xa)

    rpb = np.array(rpb)
    rub = np.array(rub)
    xb  = np.array(xb)

    nup = np.array(nup)
    ndp = np.array(ndp)

    nuq = np.array(nuq)
    ndq = np.array(ndq)

    nur = np.array(nur)
    ndr = np.array(ndr)

    nx = np.array(nx)
    ny = np.array(ny)

    return rpa,rua,xa, rpb,rub,xb, nup,ndp, nuq,ndq, nur,ndr, nx, ny

#------------------------------------------------------------------------------
def yafl_record_test_data(a, clear, noisy, t, fname):
    rp = []
    ru = []
    x  = []
    y  = []
    up = []
    dp = []
    uq = []
    dq = []
    ur = []
    dr = []
    for z in noisy:
        rp.append(a.predict())
        ru.append(a.update(z))

        x.append(a.x.copy())
        y.append(a.y.copy())

        up.append(a.Up.copy())
        dp.append(a.Dp.copy())

        uq.append(a.Uq.copy())
        dq.append(a.Dq.copy())

        ur.append(a.Ur.copy())
        dr.append(a.Dr.copy())

    rp = np.array(rp)
    ru = np.array(ru)
    x  = np.array(x)
    y  = np.array(y)
    up = np.array(up)
    dp = np.array(dp)
    uq = np.array(uq)
    dq = np.array(dq)
    ur = np.array(ur)
    dr = np.array(dr)

    with h5py.File(fname, 'w') as h5f:
        h5f.create_dataset('clear', data=clear)
        h5f.create_dataset('noisy', data=noisy)
        h5f.create_dataset('t', data=noisy)

        h5f.create_dataset('rp', data=rp)
        h5f.create_dataset('ru', data=ru)

        h5f.create_dataset('x', data=x)
        h5f.create_dataset('y', data=y)

        h5f.create_dataset('up', data=up)
        h5f.create_dataset('dp', data=dp)

        h5f.create_dataset('uq', data=uq)
        h5f.create_dataset('dq', data=dq)

        h5f.create_dataset('ur', data=ur)
        h5f.create_dataset('dr', data=dr)

#------------------------------------------------------------------------------
def yafl_file_test(b, fname):

    clear = None
    noisy = None
    t     = None

    rpa = None
    rua = None

    xa  = None
    ya  = None

    aup = None
    adp = None

    auq = None
    adq = None

    aur = None
    adr = None

    with h5py.File(fname, 'r') as h5f:
        print(h5f.keys())
        clear = h5f['clear'][:]
        noisy = h5f['noisy'][:]
        t     = h5f['t'][:]

        rpa = h5f['rp'][:]
        rua = h5f['ru'][:]

        xa  = h5f['x'][:]
        ya  = h5f['y'][:]

        aup = h5f['up'][:]
        adp = h5f['dp'][:]

        auq = h5f['uq'][:]
        adq = h5f['dq'][:]

        aur = h5f['ur'][:]
        adr = h5f['dr'][:]

    rpb = []
    rub = []
    xb  = []

    nup = []
    ndp = []
    nuq = []
    ndq = []
    nur = []
    ndr = []
    nx  = []
    ny  = []

    for i,z in enumerate(noisy):
        # Run B
        rpb.append(b.predict())
        rub.append(b.update(z))
        xb.append(b.x.copy())

        nx.append(2. * norm(xa[i] - b.x) / norm(xa[i] + b.x))
        ny.append(2. * norm(ya[i] - b.y) / norm(ya[i] + b.y))

        nup.append(2. * norm(aup[i] - b.Up) / norm(aup[i] + b.Up))
        ndp.append(2. * norm(adp[i] - b.Dp) / norm(adp[i] + b.Dp))

        nuq.append(2. * norm(auq[i] - b.Uq) / norm(auq[i] + b.Uq))
        ndq.append(2. * norm(adq[i] - b.Dq) / norm(adq[i] + b.Dq))

        nur.append(2. * norm(aur[i] - b.Ur) / norm(aur[i] + b.Ur))
        ndr.append(2. * norm(adr[i] - b.Dr) / norm(adr[i] + b.Dr))

    rpb = np.array(rpb)
    rub = np.array(rub)
    xb  = np.array(xb)

    nup = np.array(nup)
    ndp = np.array(ndp)

    nuq = np.array(nuq)
    ndq = np.array(ndq)

    nur = np.array(nur)
    ndr = np.array(ndr)

    nx = np.array(nx)
    ny = np.array(ny)

    return clear,noisy,t, rpa,rua,xa, rpb,rub,xb, nup,ndp, nuq,ndq, nur,ndr, nx, ny

#------------------------------------------------------------------------------
def filterpy_ab_test(a, b, measurements):

    xa  = []
    xb  = []

    nP = []
    nQ = []
    nR = []

    nx  = []
    ny  = []

    for z in measurements:
        # Run A
        a.predict()
        a.update(z)
        xa.append(a.x.copy())

        # Run B
        b.predict()
        b.update(z)
        xb.append(b.x.copy())

        nx.append(2. * norm(a.x - b.x) / norm(a.x + b.x))
        ny.append(2. * norm(a.y - b.y) / norm(a.y + b.y))

        nP.append(2. * norm(a.P - b.P) / norm(a.P + b.P))
        nQ.append(2. * norm(a.Q - b.Q) / norm(a.Q + b.Q))
        nR.append(2. * norm(a.R - b.R) / norm(a.R + b.R))

    xa  = np.array(xa)
    xb  = np.array(xb)

    nP = np.array(nP)
    nQ = np.array(nQ)
    nR = np.array(nR)

    nx = np.array(nx)
    ny = np.array(ny)

    return xa,xb, nP,nQ,nR, nx,ny



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
    Copyright 2021 anonimous <shkolnick-kun@gmail.com> and contributors.

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing,
    software distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.

    See the License for the specific language governing permissions
"""
import matplotlib.pyplot as plt
import numpy as np
import pyximport
import scipy.stats
import sys
import time

"""
sys.path.insert(0,'../../src/yaflpy')

pyximport.install(
    build_dir='../projects/obj',
    pyimport=True,
    reload_support=True,
    language_level=3,
    setup_args={
        'include_dirs': [np.get_include(), '../../src', '../../src/yaflpy'],
        }
    )
"""

import yaflpy
#from yaflpy import RobustJoseph as KF
#from yaflpy import RobustBierman as KF
#from yaflpy import AdaptiveRobustJoseph as KF
from yaflpy import AdaptiveRobustBierman as KF

def _fx(x, dt, **fx_args):
    x = x.copy()
    x[0] += x[1] * dt
    x[2] += x[3] * dt
    return x

def _jfx(x, dt, **fx_args):
    F = np.array([
        [1., dt, 0., 0.],
        [0., 1., 0., 0.],
        [0., 0., 1., dt],
        [0., 0., 0., 1.],
        ])
    return F

def _hx(x, **hx_args):
    if hx_args:
        print(hx_args)
    return np.array([x[0], x[2]])

def _jhx(x, **hx_args):
    H = np.array([
        [1., 0., 0., 0.],
        [0., 0., 1., 0.],
        ])
    return H

#Default noice model is Normal with Poisson outliers
def _gz(beta):
    # +- 3*sigma
    if 3.0 >= np.abs(beta):
        return float(beta)

    # +- 6*sigma - uncertain measurements
    if 6.0 >= np.abs(beta):
        return float(beta/3.0)

    # outliers
    return float(np.sign(beta))

def _gdotz(beta):
    # +- 3*sigma
    if 3.0 >= np.abs(beta):
        return 1.0

    # +- 6*sigma - uncertain measurements
    if 6.0 >= np.abs(beta):
        return 1.0/3.0

    # outliers
    return 0.0

def _zrf(a,b):
    return a - b

STD = 100.

#kf = KF(4, 2, 1., _fx, _jfx, _hx, _jhx, residual_z=_zrf)
kf = KF(4, 2, 1., _fx, _jfx, _hx, _jhx, gz=_gz, gdotz=_gdotz)
#kf = KF(4, 2, 1., _fx, _jfx, _hx, _jhx)
kf.x[0] = 1000.
kf.x[1] = -0.5
kf.Dp *= .00001
kf.Dq *= 1.0e-6
#This is robust filter, so no square here
kf.Dr *= STD*STD*0.000001
kf.Dr[0] *= .75
kf.Ur += 0.5

kf.rff = 0.001
kf.qff = 0.0001

#kf.chi2 = scipy.stats.chi2.ppf(0.996, 1)

N = 10000

clean = np.zeros((N, 2))
noisy = np.zeros((N, 2))
t     = np.zeros((N,), dtype=np.float)
st    = np.zeros((N,), dtype=np.int)
dr    = np.zeros((N,), dtype=np.int)
for i in range(1, len(clean)//2):
    clean[i] = clean[i-1] + np.array([1.5,1.])
    noisy[i] = clean[i]   + np.random.normal(scale=STD, size=2)
    t[i] = i

for i in range(i, len(clean)):
    clean[i] = clean[i-1] + np.array([1.,3.])
    noisy[i] = clean[i]   + np.random.normal(scale=STD, size=2)
    t[i] = i

kf_out = np.zeros((N, 2))

start = time.time()
for i, z in enumerate(noisy):
    kf.predict()
    st[i] = kf.update(z)
    dr[i] = np.linalg.norm(kf.Dr)
    # if st[i] & (yaflpy.ST_MSK_GLITCH_LARGE | yaflpy.ST_MSK_GLITCH_SMALL):
    #     kf.rff = 0.0
    # else:
    #     kf.rff = 0.001
    kf_out[i] = kf.x[::2]
end = time.time()
print(end - start)

plt.plot(t, noisy - kf_out)
plt.show()

plt.plot(t, clean - kf_out)
plt.show()

plt.plot(clean[:,0], clean[:,1], kf_out[:,0], kf_out[:,1])
plt.show()

plt.plot(noisy[:,0], noisy[:,1], kf_out[:,0], kf_out[:,1])
plt.show()

plt.plot(t, noisy[:,1], t, kf_out[:,1], t, clean[:,1])
plt.show()

plt.plot(t, noisy[:,0], t, kf_out[:,0], t, clean[:,0])
plt.show()

plt.plot(t, [st[i] & yaflpy.ST_MSK_ANOMALY for i in range(N)])
plt.show()

plt.plot(t, [st[i] & yaflpy.ST_MSK_REGULARIZED for i in range(N)])
plt.show()

plt.plot(t, [st[i] & yaflpy.ST_MSK_GLITCH_SMALL for i in range(N)])
plt.show()

plt.plot(t, [st[i] & yaflpy.ST_MSK_GLITCH_LARGE for i in range(N)])
plt.show()


print('Done!')



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
    Copyright 2021 anonimous <shkolnick-kun@gmail.com> and contributors.

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing,
    software distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.

    See the License for the specific language governing permissions
"""
import matplotlib.pyplot as plt
import numpy as np
import pyximport
import scipy.stats
import sys
import time

"""
sys.path.insert(0,'../../src/yaflpy')

pyximport.install(
    build_dir='../projects/obj',
    pyimport=True,
    reload_support=True,
    language_level=3,
    setup_args={
        'include_dirs' : [np.get_include(), '../../src', '../../src/yaflpy'],
        'define_macros':[("NPY_NO_DEPRECATED_API", "NPY_1_7_API_VERSION")]
        }
    )
"""
#from yaflpy import MerweSigmaPoints as SP
from yaflpy import JulierSigmaPoints as SP

from yaflpy import Unscented as KF
#from yaflpy import UnscentedAdaptive as KF
#from yaflpy import UnscentedBierman as KF
#from yaflpy import UnscentedAdaptiveBierman as KF


def _fx(x, dt, **fx_args):
    x = x.copy()
    x[0] += x[1] * dt
    x[2] += x[3] * dt
    return x


def _hx(x, **hx_args):
    if hx_args:
        print(hx_args)
    return np.array([x[0], x[2]])


def _zrf(a,b):
    #print(a - b)
    return a - b

STD = 100.

#sp = SP(4, 0.1, 2., 0)
sp = SP(4, 0.0)
kf = KF(4, 2, 1., _hx, _fx, sp)
#kf = KF(4, 2, 1., _hx, _fx, sp, residual_z=_zrf)

kf.x[0] = 0.
kf.x[1] = 0.3
kf.Dp *= .00001
kf.Dq *= 1.0e-6
#This is robust filter, so no square here
kf.Dr *= STD*STD*0.0001

kf.Dr[0] *= .75
kf.Ur += .5

#kf.chi2 = 8.807468393511947

# sp = MerweScaledSigmaPoints(4, alpha=.1, beta=2., kappa=0)
# kf = UnscentedKalmanFilter(4, 2, dt=1., fx=_fx, hx=_hx, points=sp)

# kf.P *= .1
# kf.Q *= 1e-8
# kf.R *= STD*STD

kf.rff = 0.001
kf.qff = 0.0001

print(kf.x)

N = 10000

clean = np.zeros((N, 2))
noisy = np.zeros((N, 2))
t     = np.zeros((N,), dtype=np.float)
# for i in range(1, len(clean)//2):
#     clean[i] = clean[i-1] + np.array([1.5,1.])
#     noisy[i] = clean[i]   + np.random.normal(scale=STD, size=2)
#     t[i] = i

# for i in range(i, len(clean)):
#     clean[i] = clean[i-1] + np.array([1.,1.5])
#     noisy[i] = clean[i]   + np.random.normal(scale=STD, size=2)
#     t[i] = i

for i in range(1, len(clean)):
    clean[i] = clean[i-1] + np.array([1.,1.])
    noisy[i] = clean[i]   + np.random.normal(scale=STD, size=2)
    t[i] = i

kf_out = np.zeros((N, 2))



start = time.time()
for i, z in enumerate(noisy):
    kf.predict()
    kf.update(z)
    # print('y:',kf.y)
    # print('Up:',kf.Up)
    # print('Dp:',kf.Dp)
    # print('Uq:',kf.Uq)
    # print('Dq:',kf.Dq)
    kf_out[i] = kf.zp
end = time.time()
print(end - start)



plt.plot(t, noisy - kf_out)
plt.show()

plt.plot(t, clean - kf_out)
plt.show()

plt.plot(clean[:,0], clean[:,1], kf_out[:,0], kf_out[:,1])
plt.show()

plt.plot(noisy[:,0], noisy[:,1],  kf_out[:,0], kf_out[:,1])
plt.show()

plt.plot(t, noisy[:,1], t, kf_out[:,1], t, clean[:,1])
plt.show()

plt.plot(t, noisy[:,0], t, kf_out[:,0], t, clean[:,0])
plt.show()


print('Done!')



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
    This file is based on
    https://github.com/FedericoStra/cython-package-example/blob/master/setup.py
"""
"""
    MIT License

    Copyright (c) 2019 Federico Stra

    Permission is hereby granted, free of charge, to any person obtaining a copy
    of this software and associated documentation files (the "Software"), to deal
    in the Software without restriction, including without limitation the rights
    to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
    copies of the Software, and to permit persons to whom the Software is
    furnished to do so, subject to the following conditions:

        The above copyright notice and this permission notice shall be included in all
        copies or substantial portions of the Software.

    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
    IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
    FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
    AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
    LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
    OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
    SOFTWARE.
"""
"""
    Copyright 2021 anonimous <shkolnick-kun@gmail.com> and contributors.

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing,
    software distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.

    See the License for the specific language governing permissions
"""

import numpy
import os
from os.path import dirname, isfile, join, splitext
from setuptools import Extension, setup

try:
    from Cython.Build import cythonize
    from Cython.Compiler.Main import default_options
except ImportError:
    cythonize = None

# https://cython.readthedocs.io/en/latest/src/userguide/source_files_and_compilation.html#distributing-cython-modules
def no_cythonize(extensions, **_ignore):
    for extension in extensions:
        sources = []
        for sfile in extension.sources:
            path, ext = splitext(sfile)
            if ext in (".pyx", ".py"):
                if extension.language == "c++":
                    ext = ".cpp"
                else:
                    ext = ".c"
                sfile = path + ext
            sources.append(sfile)
        extension.sources[:] = sources
    return extensions

EXT_NAME  = "yaflpy"
SETUP_DIR = dirname(__file__)
SRC_DIR   = join(SETUP_DIR, "src")
EXT_DIR   = join(SRC_DIR, EXT_NAME)

DEFINES = [("NPY_NO_DEPRECATED_API", "NPY_1_7_API_VERSION")]
CT_ENV  = {"YAFLPY_USE_64_BIT":False}

if int(os.getenv("YAFL_64", 0)):
    #Override defaults in 64 bit case
    print('YAFL: The extensuion will be compiled in 64 bit variant!!!')
    CT_ENV   = {"YAFLPY_USE_64_BIT":True}
    DEFINES += [("YAFL_USE_64_BIT",1)]

extensions = [
    Extension(EXT_NAME, [join(EXT_DIR, EXT_NAME + ".pyx")],
              language='c',
              include_dirs=[numpy.get_include(), SRC_DIR, EXT_DIR],
              define_macros=DEFINES
              )
    ]

if not isfile(join(EXT_DIR, EXT_NAME + ".c")):
    CYTHONIZE = cythonize is not None
else:
    CYTHONIZE = bool(int(os.getenv("CYTHONIZE", 0))) and cythonize is not None

if CYTHONIZE:
    default_options["compile_time_env"] = CT_ENV
    compiler_directives = {
        "language_level": 3,
        "embedsignature": True
        }
    extensions = cythonize(extensions, compiler_directives=compiler_directives)
else:
    extensions = no_cythonize(extensions)

with open("requirements.txt") as fp:
    install_requires = fp.read().strip().split("\n")

with open("requirements-dev.txt") as fp:
    dev_requires = fp.read().strip().split("\n")

setup(
    ext_modules=extensions,
    install_requires=install_requires,
    extras_require={
        "dev": dev_requires,
        "docs": ["sphinx", "sphinx-rtd-theme"]
    },
)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Thu May 21 17:08:55 2020

@author: anon
"""
import subprocess
import numpy as np
import h5py

import matplotlib.pyplot as plt

N = 10000

clear = np.zeros((N, 2))
noisy = np.zeros((N, 2))
for i in range(1, len(clear)):
    clear[i] = clear[i-1] + np.array([1.,1.])
    noisy[i] = clear[i]   + np.random.normal(scale=20., size=2)

with h5py.File('../data/input.h5', 'w') as h5f:
    h5f.create_dataset('clear', data=clear)
    h5f.create_dataset('noisy', data=noisy)

#Run kf aplication here
subprocess.call(['../projects/ada_ukf/bin/Debug/hdf5_test'])

with h5py.File('../data/output.h5', 'r') as h5f:
    kf_out = h5f['kf_out'][:]

plt.plot(clear - kf_out)
plt.show()
